'''Utility functions for low rank denoising
Copyright Will Clarke, University of Oxford, 2021'''

import numpy as np


def lsvd(A, r=None, sv_lim=None):
    """Calculate the svd decomposition of A.
    Truncated to rank r or to a SV limit (e.g. determined by MP)
    Copied from Mark Chiew's Matlab code.

    U, S, V = lsvd(A, r)

    U, V are the left and right singular vectors of A
    S contains the singular values

    A is a real or complex input matrix, where size(A,1) > size(A,2)
    r is the desired output rank of U*S*V' where 1 <= r <= size(A,2)

    :param A: Input matrix
    :type A: numpy.ndarray
    :param r: Truncate to rank r, defaults to None (no truncation)
    :type r: int, optional
    :param sv_lim: Singular value threshold, used if r is None.
        Defaults to None
    :type sv_lim: float, optional
    :return: Truncated estimate of A
    :rtype: numpy.ndarray
    """
    # Quick checks
    if r is not None and r < 1:
        raise ValueError('Rank must be > 0')
    if sv_lim is not None and sv_lim < 0.0:
        raise ValueError('sv_lim must be > 0.0')

    if A.shape[1] > A.shape[0]:
        swap = True
        A = A.conj().T
    else:
        swap = False

    D, V = np.linalg.eigh(A.conj().T @ A)
    D[D < 0.0] = 1E-15
    S = D**0.5

    if r is None and sv_lim is None:
        # No truncation
        r = A.shape[1]
    elif r is None and sv_lim > 0.0:
        r = len(S) - np.searchsorted(S, sv_lim)
        if r == 0:
            r = 1
            # raise ValueError('MP limit higher than any singular value.')
        # print(f'Truncate to r = {r}')

    S = S[-r:]
    V = V[:, -r:]
    U = A @ (V @ np.diag(1 / S))
    S = np.diag(S)

    if swap:
        return V, S, U
    else:
        return U, S, V


def svd_trunc(A, r=None, sv_lim=None):
    """SVD based truncation of A to rank r, or SV threshold sv_lim.
    A may have more than 2 dimensions, but it will be
    reshaped to two dimensions, preserving the length of the
    final dimension.

    Also returns estimates of the (unscaled) variance and covariance of the
    truncated matrix A.

    :param A: Input matrix
    :type A: numpy.ndarray
    :param r: Truncate to rank r, defaults to None (no truncation)
    :type r: int, optional
    :param sv_lim: Singular value threshold, used if r is None.
        Defaults to None
    :type sv_lim: float, optional
    :return: Truncated estimate of A
    :rtype: numpy.ndarray
    :return: Estimate of the variance in A
    :rtype: numpy.ndarray
    :return: Estimate of the covariance in the last dimension of A.
    :rtype: numpy.ndarray
    """
    dims = A.shape
    u, s, v = lsvd(A.reshape((-1, dims[-1])), r=r, sv_lim=sv_lim)

    est_var = calc_var(u, v)
    est_covar = calc_covar(v)

    return (u @ s @ v.conj().T).reshape(dims), est_var.reshape(dims), est_covar


def calc_var(u, v):
    """Calculate the unscaled variance from a truncated SVD.
    Uses the method proposed in Chen Y, Fan J, Ma C, Yan Y.
    Inference and uncertainty quantification for noisy matrix completion.
    PNAS 2019;116:22931–22937 doi: 10.1073/pnas.1910053116.

    :param u: Left singular vectors of M
    :type u: numpy.ndarray
    :param v: Right singular vectors of M
    :type v: numpy.ndarray
    :return: Estimated element-wise variance of M
    :rtype: numpy.ndarray
    """
    return (np.linalg.norm(u, axis=1)**2)[:, np.newaxis]\
        + np.linalg.norm(v, axis=1)**2


def calc_covar(u):
    """Calculate the estimated unscaled off-diagonal covariance
    from a truncated SVD.

    :param u: Left or right singular vectors of M
    :type u: numpy.ndarray
    :return: Estimated unscaled off-diagonal covariance of M
    :rtype: numpy.ndarray
    """
    return u @ u.conj().T


def max_sv_from_mp(data_var, data_shape):
    """Calculate the upper Marchenko–Pastur limit for a pure noise
    Matrix of defined shape and data variance.

    :param data_var: Noise variance
    :type data_var: float
    :param data_shape: 2-tuple of data dimensions.
    :type data_shape: tuple
    :return: Upper MP singular value limit
    :rtype: float
    """
    # Make sure dimensions agree with utils.lsvd
    # utils.lsvd will always move largest dimension to first dim.
    if data_shape[1] > data_shape[0]:
        data_shape = (data_shape[1], data_shape[0])

    c = data_shape[1] / data_shape[0]
    sv_lim = data_var * (1 + np.sqrt(c))**2
    return (sv_lim * data_shape[0])**0.5



'''Primary denoising functions for low rank denoising
Copyright Will Clarke, University of Oxford, 2021'''

import numpy as np
from itertools import combinations

from .utils import svd_trunc, lsvd, max_sv_from_mp


def st_denoising(noisy_data, rank=None, mp_var=None, patch=None, step=1):
    """Apply ST denoising truncating the data to rank r, or upto MP defined
    limit. Can be applied globally or locally (patch-wise).

    Incorporates estimation of uncertainty adapted from
    Uncertainty Quantification for Hyperspectral Image Denoising Frameworks
    based on Low-rank Matrix Approximation
    Shaobo Xia, Jingwei Song, Dong Chen, Jun Wang
    arXiv:2004.10959

    :param noisy_data:  Array of data of shape NXxNYxNZx...xNt
        (time domain last).
    :type noisy_data: np.ndarray
    :param rank: Truncate to rank r, defaults to None (no truncation). If set
        to 'MP' then truncation will be set using the limit calculated using
        the mp_var argument.
    :type rank: int or str, optional
    :param mp_var: Variance to use to estimate the MP derived truncation.
    :type mp_var: float, optional
    :param patch: 3D patch size, defaults to None for global.
    :type patch: tuple, optional
    :param step: Patch step size, defaults to 1
    :type step: int, optional
    :return: Denoised data
    :rtype: np.ndarray
    :return: Estimated variance
    :rtype: np.ndarray
    :return: Estimated covariance in final (time) dimension
    :rtype: np.ndarray
    """

    truncated_data = np.zeros_like(noisy_data)
    if patch:

        if rank is None or isinstance(rank, int):
            def svd_t(x):
                return svd_trunc(x, r=rank)
        elif isinstance(rank, str) and rank.lower() == 'mp'\
                and mp_var is not None:
            mp_shape = (np.prod(tuple(patch) + noisy_data.shape[3:-1]),
                        noisy_data.shape[-1])
            mp_lim = max_sv_from_mp(mp_var, mp_shape)

            def svd_t(x):
                return svd_trunc(x, sv_lim=mp_lim)
        else:
            raise TypeError('rank parameter must be positive integer or "mp".'
                            ' If "mp" then mp_var must be specified.')

        data_shape = noisy_data.shape[:3]
        M = np.full(data_shape, False)
        M[::step, ::step, ::step] = True
        ii, jj, kk = np.where(M)

        idx_i = np.arange(0, np.min((patch[0], data_shape[0])))
        idx_j = np.arange(0, np.min((patch[1], data_shape[1])))
        idx_k = np.arange(0, np.min((patch[2], data_shape[2])))

        navg = np.zeros((len(ii),) + data_shape)
        covar_out = []
        variance_patches = np.zeros((len(ii),) + noisy_data.shape)
        for count, (idx, jdx, kdx) in enumerate(zip(ii, jj, kk)):
            i = np.mod(idx+idx_i-1, data_shape[0])
            j = np.mod(jdx+idx_j-1, data_shape[1])
            k = np.mod(kdx+idx_k-1, data_shape[2])

            denoised, var, cvar = svd_t(noisy_data[np.ix_(i, j, k)])
            truncated_data[np.ix_(i, j, k)] += denoised
            variance_patches[np.ix_([count], i, j, k)] = var
            covar_out.append(cvar)
            navg[np.ix_([count], i, j, k)] += 1

        navg_all = np.sum(navg, axis=0)
        # Deal with broadcasting comparing along trailing dimension
        truncated_data = (truncated_data.T / navg_all.T).T

        variance = calculate_patch_variance(variance_patches, navg, patch)

        covar = np.asarray(covar_out).mean(axis=0)
    else:
        # Default to global
        if rank is None or isinstance(rank, int):
            truncated_data, variance, covar = svd_trunc(noisy_data, r=rank)
        elif isinstance(rank, str) and rank.lower() == 'mp'\
                and mp_var is not None:
            mp_shape = (np.prod(noisy_data.shape[:-1]), noisy_data.shape[-1])
            mp_lim = max_sv_from_mp(mp_var, mp_shape)
            truncated_data, variance, covar = svd_trunc(noisy_data,
                                                        sv_lim=mp_lim)
        else:
            raise TypeError('rank parameter must be positive integer or "mp".'
                            ' If "mp" then mp_var must be specified.')

    return truncated_data, variance, covar


def calculate_patch_variance(patch_var, patch_positions, patch_size):
    """Function to calculate patch-wise variance

    Based on Uncertainty Quantification for Hyperspectral Image Denoising
    Frameworks based on Low-rank Matrix Approximation
    Shaobo Xia, Jingwei Song, Dong Chen, Jun Wang
    arXiv:2004.10959

    :param patch_var: Array of variance with first dimension
        containing patches
    :type patch_var: numpy.ndarray
    :param patch_positions: Array of patch positions with first
        dimension containing patches
    :type patch_positions: numpy.ndarray
    :param patch_size: 3D patch size
    :type patch_size: tuple
    :return: Estimated variance
    :rtype: numpy.ndarray
    """
    # Number of patches in each position
    npatch_all = np.sum(patch_positions, axis=0)

    # Calculate the first (self) term of the variance
    var_self = (np.sum(patch_var, axis=0).T / npatch_all.T**2).T

    # Calculate the cross term
    # Calculate the overlap of each patch with each other
    # Equivalent to prameter nu^{(ijk)(i'j'k')}
    overlap_mat = np.zeros((len(patch_positions), len(patch_positions)))
    for idx, ppi in enumerate(patch_positions):
        for jdx, ppj in enumerate(patch_positions):
            overlap_mat[idx, jdx] = np.sum(ppi * ppj) / np.prod(patch_size)

    # Loop over all voxels and sum up the cross terms
    var_cross = np.zeros_like(patch_var[0])
    for ind in np.ndindex(npatch_all.shape):
        vox_patches = patch_positions[:, ind[0], ind[1], ind[2]].nonzero()[0]
        for comb in combinations(vox_patches, 2):
            var_cross[ind[0], ind[1], ind[2], ...] += \
                2 * overlap_mat[comb[0], comb[1]]\
                * np.sqrt(patch_var[comb[0], ind[0], ind[1], ind[2], ...])\
                * np.sqrt(patch_var[comb[1], ind[0], ind[1], ind[2], ...])

    # Normalise the cross terms
    var_cross = (var_cross.T / (npatch_all.T)**2).T

    return var_self + var_cross


def lp(data, r,  W=None):
    '''Run linear-prediction denoising on 3D MRSI data.
    :param np.ndarray data: space-time MRSI data
                      assume dimensions [Nx, Ny, Nz, ..., Nt]
    :param int r: rank threshold for Hankel filtering constraint
    :param int W: Size of 1st Hankel matrix dimension.
        If None, defaults to half of time domain length.
    '''
    data_shape = data.shape
    out = data.copy().reshape(-1, data_shape[-1])

    # Loop over all the FIDs and apply
    # Hankel matrix Low-Rank enforcement
    if W is None:
        W = int(out.shape[1] / 2)
    elif W >= out.shape[1]:
        raise ValueError('W must be smaller than the time domain length.')
    K = out.shape[1] - W
    H = np.zeros((W, W+1), dtype=complex)
    for idx, v_data in enumerate(out):
        for wdx in range(W):
            H[wdx, :] = v_data[wdx:(wdx+K+1)]
        H, _, _ = svd_trunc(H, r=r)

        out[idx, :] = np.concatenate((H[0, :], H[1:, -1].T))

    return out.reshape(data_shape)


def lora(data, r1, r2, r1_mp_var=None, W=None):
    '''Run LORA denoising on 3D MRSI data.
    :param np.ndarray data: space-time MRSI data
                      assume dimensions [Nx, Ny, Nz, ..., Nt]
    :param int r1: rank threshold for space-time LR constraint.
        Set to 'mp' for Marchenko–Pastur threshold.
    :param int r2: rank threshold for Hankel filtering constraint
    :param float r1_mp_var: Variance of noise in input data
    :param int W: Size of 1st Hankel matrix dimension.
        If None, defaults to half of time domain length.
    '''

    if r1 is None or isinstance(r1, int):
        def svd_t(x):
            return svd_trunc(x, r=r1)
    elif isinstance(r1, str) and r1.lower() == 'mp' and r1_mp_var is not None:
        mp_shape = (np.prod(data.shape[:-1]), data.shape[-1])
        mp_lim = max_sv_from_mp(r1_mp_var, mp_shape)

        def svd_t(x):
            return svd_trunc(x, sv_lim=mp_lim)
    else:
        raise TypeError('r1 parameter must be positive integer or "mp".'
                        ' If "mp" then r1_mp_var must be specified.')

    data_shape = data.shape
    out = data.copy().reshape(-1, data_shape[-1])

    # Space-time low rank enforcement
    out = out.T
    out, _, _ = svd_t(out)
    out = out.T

    # Loop over all the FIDs and apply
    # Hankel matrix Low-Rank enforcement
    if W is None:
        W = int(out.shape[1] / 2)
    elif W >= out.shape[1]:
        raise ValueError('W must be smaller than the time domain length.')
    K = out.shape[1] - W
    H = np.zeros((W, W+1), dtype=complex)
    for idx, v_data in enumerate(out):
        for wdx in range(W):
            H[wdx, :] = v_data[wdx:(wdx+K+1)]
        H, _, _ = svd_trunc(H, r=r2)

        out[idx, :] = np.concatenate((H[0, :], H[1:, -1].T))

    return out.reshape(data_shape)


def llora(data, r1, r2, patch, step=1, r1_mp_var=None, W=None):
    '''Run local-LORA denoising on 3D MRSI data.
    :param np.ndarray data: space-time MRSI data
                      assume dimensions [Nx, Ny, Nz, ..., Nt]
    :param int r1: rank threshold for space-time LR constraint
    :param int r2: rank threshold for Hankel filtering constraint
    :param patch: Spatial patch size, 3-tuple of patch in X,Y,Z.
    :param step: Size of patch step.
    :param float r1_mp_var: Variance of noise in input data
    :param int W: Size of 1st Hankel matrix dimension.
        If None, defaults to half of time domain length.
    '''
    data_shape = data.shape[:3]
    M = np.full(data_shape, False)
    M[::step, ::step, ::step] = True
    ii, jj, kk = np.where(M)

    idx_i = np.arange(0, np.min((patch[0], data_shape[0])))
    idx_j = np.arange(0, np.min((patch[1], data_shape[1])))
    idx_k = np.arange(0, np.min((patch[2], data_shape[2])))

    navg = np.zeros((len(ii),) + data_shape)
    out = np.zeros_like(data)
    for count, (idx, jdx, kdx) in enumerate(zip(ii, jj, kk)):
        i = np.mod(idx+idx_i-1, data_shape[0])
        j = np.mod(jdx+idx_j-1, data_shape[1])
        k = np.mod(kdx+idx_k-1, data_shape[2])

        out[np.ix_(i, j, k)] += lora(data[np.ix_(i, j, k)],
                                     r1,
                                     r2,
                                     r1_mp_var=r1_mp_var,
                                     W=W)

        navg[np.ix_([count], i, j, k)] += 1

    navg_all = np.sum(navg, axis=0)
    # Deal with broadcasting comparing along trailing dimension
    out = (out.T / navg_all.T).T

    return out


def sure_svt(data, var, patch, step=1):
    '''Run SURE optimised SVT denoising on 3D MRSI data.
    :param np.ndarray data: space-time MRSI data
                      assume dimensions [Nx, Ny, Nz, ..., Nt]
    :param float var: Variance of real / imag noise in input data.
         I.e. np.var(complex_noise)/2 == np.var(np.real(complex_noise))
    :param patch: Spatial patch size, 3-tuple of patch in X,Y,Z.
    :param step: Size of patch step.
    '''
    def div_SVT(lam, S, Nx, Nt):
        z = (2 * np.abs(Nx - Nt) + 1)\
            * np.sum(np.maximum(1 - lam / S, 0))\
            + np.sum(S > lam)

        ss = np.subtract.outer(S**2, S**2)
        ss = ss[~np.eye(S.shape[0], dtype=bool)].reshape(S.shape[0], -1)
        z += np.sum(4 * S * np.maximum(S - lam, 0) / ss.T)

        return z

    def SURE(lam, S, v, Nx, Nt):
        return -2 * Nx * Nt * v\
            + np.sum(np.minimum(S**2, lam**2))\
            + 2 * v * div_SVT(lam, S, Nx, Nt)

    data_shape = data.shape[:3]
    M = np.full(data_shape, False)
    M[::step, ::step, ::step] = True
    ii, jj, kk = np.where(M)

    idx_i = np.arange(0, np.min((patch[0], data_shape[0])))
    idx_j = np.arange(0, np.min((patch[1], data_shape[1])))
    idx_k = np.arange(0, np.min((patch[2], data_shape[2])))

    navg = np.zeros((len(ii),) + data_shape)
    out = np.zeros_like(data)
    for count, (idx, jdx, kdx) in enumerate(zip(ii, jj, kk)):
        i = np.mod(idx+idx_i-1, data_shape[0])
        j = np.mod(jdx+idx_j-1, data_shape[1])
        k = np.mod(kdx+idx_k-1, data_shape[2])

        curr_data = data[np.ix_(i, j, k)]\
            .reshape(-1, data[np.ix_(i, j, k)].shape[-1])

        U, S, V = lsvd(curr_data)
        S = np.diag(S)

        # Search up to second highest singular value.
        q = np.logspace(np.log10(S[-2]), np.log10(S[0]), num=50)
        w = np.zeros((50,))

        for idx in range(50):
            w[idx] = SURE(q[idx],
                          S,
                          var,
                          curr_data.shape[0],
                          curr_data.shape[-1])

        thresh_SURE = q[np.argmin(w)]
        # breakpoint()
        S_scaled = np.diag(np.maximum(S-thresh_SURE, 0)
                           * S[-1] / (S[-1] - thresh_SURE))
        out[np.ix_(i, j, k)] += \
            (U @ S_scaled @ V.conj().T)\
            .reshape(data[np.ix_(i, j, k)].shape)

        navg[np.ix_([count], i, j, k)] += 1

    navg_all = np.sum(navg, axis=0)
    # Deal with broadcasting comparing along trailing dimension
    out = (out.T / navg_all.T).T

    return out


def sure_hard(data, var, patch, step=1):
    '''Run SURE optimised SVT denoising on 3D MRSI data.
    :param np.ndarray data: space-time MRSI data
                      assume dimensions [Nx, Ny, Nz, ..., Nt]
    :param float var: Variance of real / imag noise in input data.
         I.e. np.var(complex_noise)/2 == np.var(np.real(complex_noise))
    :param patch: Spatial patch size, 3-tuple of patch in X,Y,Z.
    :param step: Size of patch step.
    '''
    def div_HARD(r, S, Nx, Nt):
        z = (Nx + Nt) * r - r**2
        for idx in range(r):
            for jdx in range(r, S.size):
                z += 2*S[jdx]**2 / (S[idx]**2 - S[jdx]**2)
        return z

    def SURE(r, S, v, Nx, Nt):
        return -2 * Nx * Nt * v \
               + np.sum(S[r:]**2) \
               + 4 * v * div_HARD(r, S, Nx, Nt)

    data_shape = data.shape[:3]
    M = np.full(data_shape, False)
    M[::step, ::step, ::step] = True
    ii, jj, kk = np.where(M)

    idx_i = np.arange(0, np.min((patch[0], data_shape[0])))
    idx_j = np.arange(0, np.min((patch[1], data_shape[1])))
    idx_k = np.arange(0, np.min((patch[2], data_shape[2])))

    navg = np.zeros((len(ii),) + data_shape)
    out = np.zeros_like(data)
    for count, (idx, jdx, kdx) in enumerate(zip(ii, jj, kk)):
        i = np.mod(idx+idx_i-1, data_shape[0])
        j = np.mod(jdx+idx_j-1, data_shape[1])
        k = np.mod(kdx+idx_k-1, data_shape[2])

        curr_data = data[np.ix_(i, j, k)]\
            .reshape(-1, data[np.ix_(i, j, k)].shape[-1])

        U, S, V = lsvd(curr_data)
        s = np.diag(S)

        sure = np.zeros((curr_data.shape[0],))
        rank = np.arange(1, curr_data.shape[0]+1)

        for r in rank:
            sure[r-1] = SURE(r,
                             s[::-1],
                             var,
                             curr_data.shape[0],
                             curr_data.shape[-1])

        thresh_SURE = rank[np.argmin(sure)]
        # print(f'SURE hard = {thresh_SURE}')
        # breakpoint()

        S = S[-thresh_SURE:, -thresh_SURE:]
        V = V[:, -thresh_SURE:]
        U = U[:, -thresh_SURE:]

        out[np.ix_(i, j, k)] += \
            (U @ S @ V.conj().T)\
            .reshape(data[np.ix_(i, j, k)].shape)

        navg[np.ix_([count], i, j, k)] += 1

    navg_all = np.sum(navg, axis=0)
    # Deal with broadcasting comparing along trailing dimension
    out = (out.T / navg_all.T).T

    return out



"""Wrapper script for the mrsi denoising utilities
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright Will Clarke, University of Oxford, 2021"""

import argparse
import sys

import nibabel as nib
from pathlib import Path
import numpy as np

import mrs_denoising.denoising as denoise


def main():
    parser = argparse.ArgumentParser(description='Low-rank denoising of MRSI data.')

    def add_common_parameters(subparser):
        subparser.add_argument(
            'input', type=Path, help='Input MRSI NIfTI file.'
            ' Must have four dimensions and have time domain as 4th dimension')
        subparser.add_argument('output', type=Path, help='Name and location of output')
        subparser.add_argument(
            '--mask', type=Path,
            help='Spatial mask (NIfTI format), must be spatially contiguous and cuboidal.')
        return subparser

    subparsers = parser.add_subparsers(title='algorithms',
                                       description='Different denoising options')

    # st_denoising
    parser_st = subparsers.add_parser('st', help='Spatio-temporal denoising.')
    parser_st = add_common_parameters(parser_st)
    parser_st.add_argument('noise', nargs='+',
                           help='Either indicies (low, high) of noise region in time domain'
                                ' or estimate of noise variance (single float).')
    st_me = parser_st.add_mutually_exclusive_group()
    st_me.add_argument('-r', '--rank', type=int, help='Truncate to rank r.')
    st_me.add_argument('-mp', '--marchenko_pastur', action="store_true", help='Truncate based on MP threshold.')
    parser_st.add_argument('-p', '--patch', type=int, nargs=3, default=None,
                           help='3D patch size e.g. <3, 3, 3> or <3, 1, 1>')
    parser_st.add_argument('-s', '--step', type=int, default=1,
                           help='Patch step, default = 1. Only active if "-p/--patch" set.')
    parser_st.set_defaults(func=st_denoise)

    # lp_denoising
    parser_lp = subparsers.add_parser('lp', help='Linear-predictability denoising.')
    parser_lp = add_common_parameters(parser_lp)
    parser_lp.add_argument('rank', type=int, help='Truncate Hankel matrix to rank r')
    parser_lp.add_argument('-w', '--hankel_w', type=int,
                           help='Size of Hankel matrix first dimension. Defaults to half.')
    parser_lp.set_defaults(func=lp_denoise)

    # lora_denoising
    parser_lo = subparsers.add_parser('lora', help='(local)LORA denoising.')
    parser_lo = add_common_parameters(parser_lo)
    lo_me = parser_lo.add_mutually_exclusive_group()
    lo_me.add_argument('-r1', '--rank1', type=int, help='Truncate ST to rank r1')
    lo_me.add_argument('-mp', '--marchenko_pastur', action="store_true",
                       help='ST truncation based on MP threshold. Must also set noise argument.')
    parser_lo.add_argument('-n', '--noise', nargs='+',
                           help='Either indicies (low, high) of noise region in time domain'
                                ' or estimate of noise variance (single float).')
    parser_lo.add_argument('r2', type=int, help='Truncate Hankel matrix to rank r')
    parser_lo.add_argument('-w', '--hankel_w', type=int,
                           help='Size of Hankel matrix first dimension. Defaults to half.')
    parser_lo.add_argument('-p', '--patch', type=int, nargs=3, default=None,
                           help='3D patch size e.g. <3, 3, 3> or <3, 1, 1>')
    parser_lo.add_argument('-s', '--step', type=int, default=1,
                           help='Patch step, default = 1. Only active if "-p/--patch" set.')
    parser_lo.set_defaults(func=lora_denoise)

    # svt_denoising
    parser_svt = subparsers.add_parser('svt', help='SURE optimised soft thresholding (SVT).')
    parser_svt = add_common_parameters(parser_svt)
    parser_svt.add_argument('noise', nargs='+',
                            help='Either indicies (low, high) of noise region in time domain'
                                 ' or estimate of noise variance (single float).')
    parser_svt.add_argument('-p', '--patch', type=int, nargs=3, required=True,
                            help='3D patch size e.g. <3, 3, 3> or <3, 1, 1>')
    parser_svt.add_argument('-s', '--step', type=int, default=1,
                            help='Patch step, default = 1.')
    parser_svt.set_defaults(func=sure_svt)

    # svht_denoising
    parser_svht = subparsers.add_parser('svht', help='SURE optimised hard thresholding (SVHT).')
    parser_svht = add_common_parameters(parser_svht)
    parser_svht.add_argument('noise', nargs='+',
                             help='Either indicies (low, high) of noise region in time domain'
                                  ' or estimate of noise variance (single float).')
    parser_svht.add_argument('-p', '--patch', type=int, nargs=3, required=True,
                             help='3D patch size e.g. <3, 3, 3> or <3, 1, 1>')
    parser_svht.add_argument('-s', '--step', type=int, default=1,
                             help='Patch step, default = 1.')
    parser_svht.set_defaults(func=sure_svht)

    args = parser.parse_args()
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args.func(args)


def st_denoise(args):
    noisy_data = nib.load(args.input)
    masked_data = mask_data(noisy_data, args)

    if len(args.noise) == 2:
        noise_var = np.var(masked_data[:, :, :, int(args.noise[0]):int(args.noise[1])])
    elif len(args.noise) == 1:
        noise_var = args.noise
    else:
        raise TypeError('Noise argument must be either be single float (variance) or two ints (indices).')

    if args.marchenko_pastur:
        mpvar = noise_var
        rank = 'MP'
    else:
        mpvar = None
        rank = args.rank

    denoised, est_var, est_covar = denoise.st_denoising(masked_data,
                                                        rank=rank,
                                                        mp_var=mpvar,
                                                        patch=args.patch,
                                                        step=args.step)

    full_denoised = unmask_data(denoised, noisy_data, args)
    full_var = unmask_data(est_var, noisy_data, args)

    # Scale the (co)variance
    full_var *= noise_var
    est_covar *= noise_var

    denoised_img = nib.Nifti2Image(full_denoised, noisy_data.affine, noisy_data.header)
    var_img = nib.Nifti2Image(full_var, noisy_data.affine, noisy_data.header)

    # Save output
    nib.save(denoised_img, args.output)
    var_path = args.output.parent / (str(args.output.with_suffix('').with_suffix('').stem) + "_var.nii.gz")
    nib.save(var_img, var_path)
    covar_path = args.output.parent / (str(args.output.with_suffix('').with_suffix('').stem) + "_covar")
    np.savetxt(covar_path, est_covar)


# lp(data, r,  W=None)
def lp_denoise(args):
    noisy_data = nib.load(args.input)
    masked_data = mask_data(noisy_data, args)

    denoised = denoise.lp(masked_data,
                          rank=args.rank,
                          w=args.hankel_w)

    full_denoised = unmask_data(denoised, noisy_data, args)
    denoised_img = nib.Nifti2Image(full_denoised, noisy_data.affine, noisy_data.header)
    nib.save(denoised_img, args.output)


# lora(data, r1, r2, r1_mp_var=None, W=None)
# llora(data, r1, r2, patch, step=1, r1_mp_var=None, W=None)
def lora_denoise(args):
    noisy_data = nib.load(args.input)
    masked_data = mask_data(noisy_data, args)

    if args.marchenko_pastur:
        r1 = 'MP'
    else:
        r1 = args.rank1

    if args.noise and len(args.noise) == 2:
        noise_var = np.var(masked_data[:, :, :, int(args.noise[0]):int(args.noise[1])])
    elif args.noise and len(args.noise) == 1:
        noise_var = args.noise
    elif args.noise is None:
        noise_var = None
    else:
        raise TypeError('Noise argument must be either be single float (variance) or two ints (indices).')

    if args.patch:
        denoised = denoise.llora(masked_data,
                                 r1,
                                 args.r2,
                                 args.patch,
                                 r1_mp_var=noise_var,
                                 step=args.step,
                                 W=args.hankel_w)
    else:
        denoised = denoise.lora(masked_data,
                                r1,
                                args.r2,
                                r1_mp_var=None,
                                W=args.hankel_w)

    full_denoised = unmask_data(denoised, noisy_data, args)
    denoised_img = nib.Nifti2Image(full_denoised, noisy_data.affine, noisy_data.header)
    nib.save(denoised_img, args.output)


# sure_svt(data, var, patch, step=1):
def sure_svt(args):
    noisy_data = nib.load(args.input)
    masked_data = mask_data(noisy_data, args)

    if len(args.noise) == 2:
        noise_var = np.var(masked_data[:, :, :, int(args.noise[0]):int(args.noise[1])])
    elif len(args.noise) == 1:
        noise_var = args.noise
    else:
        raise TypeError('Noise argument must be either be single float (variance) or two ints (indices).')

    denoised = denoise.sure_svt(masked_data,
                                noise_var,
                                args.patch,
                                step=args.step)

    full_denoised = unmask_data(denoised, noisy_data, args)
    denoised_img = nib.Nifti2Image(full_denoised, noisy_data.affine, noisy_data.header)
    nib.save(denoised_img, args.output)


# sure_hard(data, var, patch, step=1)
def sure_svht(args):
    noisy_data = nib.load(args.input)
    masked_data = mask_data(noisy_data, args)

    if len(args.noise) == 2:
        noise_var = np.var(masked_data[:, :, :, int(args.noise[0]):int(args.noise[1])])
    elif len(args.noise) == 1:
        noise_var = args.noise
    else:
        raise TypeError('Noise argument must be either be single float (variance) or two ints (indices).')

    denoised = denoise.sure_hard(masked_data,
                                 noise_var,
                                 args.patch,
                                 step=args.step)

    full_denoised = unmask_data(denoised, noisy_data, args)
    denoised_img = nib.Nifti2Image(full_denoised, noisy_data.affine, noisy_data.header)
    nib.save(denoised_img, args.output)


def mask_data(img, args):
    """Convert full size data to masked data

    :param img: Nibabel nifti image
    :param args: The interpreted script arguments
    :return: Masked data
    :rtype: numpy.ndarray
    """
    full_data = img.get_fdata(dtype=complex)
    if args.mask:
        cm, rm, zm = process_mask(args)
        return full_data[cm, rm, zm, :]
    else:
        return full_data


def unmask_data(denoised, img, args):
    """Convert masked data to full size data

    :param denoised: Denoised numpy.ndarray
    :param img: Nibabel nifti image
    :param args: The interpreted script arguments
    :return: Denoised data input into fullsize array
    """
    if args.mask:
        cm, rm, zm = process_mask(args)
        full = np.zeros(img.shape, dtype=complex)
        full[cm, rm, zm, :] = denoised
        return full
    else:
        return denoised


def process_mask(args):
    '''Get bounding box of mask'''
    mask = nib.load(args.mask).get_fdata()
    mask = np.atleast_3d(mask)
    cmm, rmm, zmm = bbox_3D(mask)
    return slice(cmm[0], cmm[1]), slice(rmm[0], rmm[1]), slice(zmm[0], zmm[1])


def bbox_3D(img):
    """Get bounds of masking array

    :param img: 3D masking array
    :type img: numpy.ndarray
    :return: Three tuples of (min, max)
    """
    r = np.any(img, axis=(1, 2))
    c = np.any(img, axis=(0, 2))
    z = np.any(img, axis=(0, 1))

    rmin, rmax = np.where(r)[0][[0, -1]]
    cmin, cmax = np.where(c)[0][[0, -1]]
    zmin, zmax = np.where(z)[0][[0, -1]]

    return (rmin, rmax + 1), (cmin, cmax + 1), (zmin, zmax + 1)


if __name__ == '__main__':
    main()
